from textwrap import dedent
from typing import Any, Dict, List, Optional, Union

from ..language import DirectiveLocation

try:
    from typing import Literal, TypedDict
except ImportError:  # Python < 3.8
    from typing_extensions import Literal, TypedDict

__all__ = [
    "get_introspection_query",
    "IntrospectionDirective",
    "IntrospectionEnumType",
    "IntrospectionField",
    "IntrospectionInputObjectType",
    "IntrospectionInputValue",
    "IntrospectionInterfaceType",
    "IntrospectionListType",
    "IntrospectionNonNullType",
    "IntrospectionObjectType",
    "IntrospectionQuery",
    "IntrospectionScalarType",
    "IntrospectionSchema",
    "IntrospectionType",
    "IntrospectionTypeRef",
    "IntrospectionUnionType",
]


def get_introspection_query(
    descriptions: bool = True,
    specified_by_url: bool = False,
    directive_is_repeatable: bool = False,
    schema_description: bool = False,
    input_value_deprecation: bool = False,
    experimental_directive_deprecation: bool = False,
    input_object_one_of: bool = False,
    type_depth: int = 9,
) -> str:
    """Get a query for introspection.

    Optionally, you can exclude descriptions, include specification URLs,
    include repeatability of directives, and specify whether to include
    the schema description as well.

    The ``type_depth`` argument controls how deep to recurse into nested types.
    Larger values will result in more accurate results, but have a higher load
    on the server. Some servers might restrict the maximum query depth or
    complexity. If that's the case, try decreasing this value. The default is 9.
    """
    maybe_description = "description" if descriptions else ""
    maybe_specified_by_url = "specifiedByURL" if specified_by_url else ""
    maybe_directive_is_repeatable = "isRepeatable" if directive_is_repeatable else ""
    maybe_schema_description = maybe_description if schema_description else ""
    maybe_input_object_one_of = "isOneOf" if input_object_one_of else ""

    def input_deprecation(string: str) -> Optional[str]:
        return string if input_value_deprecation else ""

    def directive_deprecation(string: str) -> Optional[str]:
        return string if experimental_directive_deprecation else ""

    def of_type(level: int, indent: str) -> str:
        if level <= 0:
            return ""
        if level > 100:
            msg = (
                "Please set type_depth to a reasonable value"
                " between 0 and 100; the default is 9."
            )
            raise ValueError(msg)
        return (
            f"\n{indent}ofType {{"
            f"\n{indent}  name"
            f"\n{indent}  kind{of_type(level - 1, indent + '  ')}"
            f"\n{indent}}}"
        )

    return dedent(f"""
        query IntrospectionQuery {{
          __schema {{
            {maybe_schema_description}
            queryType {{ name kind }}
            mutationType {{ name kind }}
            subscriptionType {{ name kind }}
            types {{
              ...FullType
            }}
            directives{directive_deprecation("(includeDeprecated: true)")} {{
              name
              {maybe_description}
              {maybe_directive_is_repeatable}
              {directive_deprecation("isDeprecated")}
              {directive_deprecation("deprecationReason")}
              locations
              args{input_deprecation("(includeDeprecated: true)")} {{
                ...InputValue
              }}
            }}
          }}
        }}

        fragment FullType on __Type {{
          kind
          name
          {maybe_description}
          {maybe_specified_by_url}
          {maybe_input_object_one_of}
          fields(includeDeprecated: true) {{
            name
            {maybe_description}
            args{input_deprecation("(includeDeprecated: true)")} {{
              ...InputValue
            }}
            type {{
              ...TypeRef
            }}
            isDeprecated
            deprecationReason
          }}
          inputFields{input_deprecation("(includeDeprecated: true)")} {{
            ...InputValue
          }}
          interfaces {{
            ...TypeRef
          }}
          enumValues(includeDeprecated: true) {{
            name
            {maybe_description}
            isDeprecated
            deprecationReason
          }}
          possibleTypes {{
            ...TypeRef
          }}
        }}

        fragment InputValue on __InputValue {{
          name
          {maybe_description}
          type {{ ...TypeRef }}
          defaultValue
          {input_deprecation("isDeprecated")}
          {input_deprecation("deprecationReason")}
        }}

        fragment TypeRef on __Type {{
          kind
          name{of_type(type_depth, "          ")}
        }}
        """)


# Unfortunately, the following type definitions are a bit simplistic
# because of current restrictions in the typing system (mypy):
# - no recursion, see https://github.com/python/mypy/issues/731
# - no generic typed dicts, see https://github.com/python/mypy/issues/3863

# simplified IntrospectionNamedType to avoids cycles
SimpleIntrospectionType = Dict[str, Any]


class MaybeWithDescription(TypedDict, total=False):
    description: Optional[str]


class WithName(MaybeWithDescription):
    name: str


class MaybeWithSpecifiedByUrl(TypedDict, total=False):
    specifiedByURL: Optional[str]


class WithDeprecated(TypedDict):
    isDeprecated: bool
    deprecationReason: Optional[str]


class MaybeWithDeprecated(TypedDict, total=False):
    isDeprecated: bool
    deprecationReason: Optional[str]


class IntrospectionInputValue(WithName, MaybeWithDeprecated):
    type: SimpleIntrospectionType  # should be IntrospectionInputType
    defaultValue: Optional[str]


class IntrospectionField(WithName, WithDeprecated):
    args: List[IntrospectionInputValue]
    type: SimpleIntrospectionType  # should be IntrospectionOutputType


class IntrospectionEnumValue(WithName, WithDeprecated):
    pass


class MaybeWithIsRepeatable(TypedDict, total=False):
    isRepeatable: bool


class IntrospectionDirective(WithName, MaybeWithIsRepeatable, MaybeWithDeprecated):
    locations: List[DirectiveLocation]
    args: List[IntrospectionInputValue]


class IntrospectionScalarType(WithName, MaybeWithSpecifiedByUrl):
    kind: Literal["scalar"]


class IntrospectionInterfaceType(WithName):
    kind: Literal["interface"]
    fields: List[IntrospectionField]
    interfaces: List[SimpleIntrospectionType]  # should be InterfaceType
    possibleTypes: List[SimpleIntrospectionType]  # should be NamedType


class IntrospectionObjectType(WithName):
    kind: Literal["object"]
    fields: List[IntrospectionField]
    interfaces: List[SimpleIntrospectionType]  # should be InterfaceType


class IntrospectionUnionType(WithName):
    kind: Literal["union"]
    possibleTypes: List[SimpleIntrospectionType]  # should be NamedType


class IntrospectionEnumType(WithName):
    kind: Literal["enum"]
    enumValues: List[IntrospectionEnumValue]


class IntrospectionInputObjectType(WithName):
    kind: Literal["input_object"]
    inputFields: List[IntrospectionInputValue]
    isOneOf: bool


IntrospectionType = Union[
    IntrospectionScalarType,
    IntrospectionObjectType,
    IntrospectionInterfaceType,
    IntrospectionUnionType,
    IntrospectionEnumType,
    IntrospectionInputObjectType,
]

IntrospectionOutputType = Union[
    IntrospectionScalarType,
    IntrospectionObjectType,
    IntrospectionInterfaceType,
    IntrospectionUnionType,
    IntrospectionEnumType,
]

IntrospectionInputType = Union[
    IntrospectionScalarType, IntrospectionEnumType, IntrospectionInputObjectType
]


class IntrospectionListType(TypedDict):
    kind: Literal["list"]
    ofType: SimpleIntrospectionType  # should be IntrospectionType


class IntrospectionNonNullType(TypedDict):
    kind: Literal["non_null"]
    ofType: SimpleIntrospectionType  # should be IntrospectionType


IntrospectionTypeRef = Union[
    IntrospectionType, IntrospectionListType, IntrospectionNonNullType
]


class IntrospectionSchema(MaybeWithDescription):
    queryType: IntrospectionObjectType
    mutationType: Optional[IntrospectionObjectType]
    subscriptionType: Optional[IntrospectionObjectType]
    types: List[IntrospectionType]
    directives: List[IntrospectionDirective]


# The root typed dictionary for schema introspections.
# Note: We don't use class syntax here since the key looks like a private attribute.
IntrospectionQuery = TypedDict(
    "IntrospectionQuery",
    {"__schema": IntrospectionSchema},
)



from ..error import GraphQLSyntaxError
from .ast import Token
from .character_classes import is_name_start
from .lexer import Lexer
from .token_kind import TokenKind

__all__ = ["SchemaCoordinateLexer"]


_KIND_FOR_PUNCT = {
    ".": TokenKind.DOT,
    "(": TokenKind.PAREN_L,
    ")": TokenKind.PAREN_R,
    ":": TokenKind.COLON,
    "@": TokenKind.AT,
}


class SchemaCoordinateLexer(Lexer):
    """GraphQL Schema Coordinate Lexer

    A SchemaCoordinateLexer is a stateful stream generator in that every time it is
    advanced, it returns the next token in the Source. Assuming the source lexes, the
    final Token emitted by the lexer will be of kind EOF, after which the lexer will
    repeatedly return the same EOF token whenever called.

    Unlike the regular Lexer, this lexer uses a restricted syntax that does not allow
    any ignored tokens (such as whitespace or comments). Since a schema coordinate may
    not contain a newline, the line is always 1 and the line start is always 0.
    """

    def read_next_token(self, start: int) -> Token:
        """Get the next token from the source starting at the given position.

        This lexes punctuators and names only, raising a syntax error on any other
        character (including ignored tokens such as whitespace and comments).
        """
        body = self.source.body
        body_length = len(body)
        position = start

        if position < body_length:
            char = body[position]

            kind = _KIND_FOR_PUNCT.get(char)
            if kind:
                return self.create_token(kind, position, position + 1)

            if is_name_start(char):
                return self.read_name(position)

            raise GraphQLSyntaxError(
                self.source,
                position,
                f"Invalid character: {self.print_code_point_at(position)}.",
            )

        return self.create_token(TokenKind.EOF, body_length, body_length)



from typing import NamedTuple, Optional, Union

from ..language import (
    ArgumentCoordinateNode,
    DirectiveArgumentCoordinateNode,
    DirectiveCoordinateNode,
    MemberCoordinateNode,
    SchemaCoordinateNode,
    Source,
    TypeCoordinateNode,
    parse_schema_coordinate,
)
from ..pyutils import inspect
from ..type import (
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLSchema,
)

__all__ = [
    "resolve_schema_coordinate",
    "resolve_ast_schema_coordinate",
    "ResolvedNamedType",
    "ResolvedField",
    "ResolvedInputField",
    "ResolvedEnumValue",
    "ResolvedFieldArgument",
    "ResolvedDirective",
    "ResolvedDirectiveArgument",
    "ResolvedSchemaElement",
]


class ResolvedNamedType(NamedTuple):
    """A named type resolved from a schema coordinate."""

    type: GraphQLNamedType
    kind: str = "NamedType"


class ResolvedField(NamedTuple):
    """A field resolved from a schema coordinate."""

    type: Union[GraphQLObjectType, GraphQLInterfaceType]
    field: GraphQLField
    kind: str = "Field"


class ResolvedInputField(NamedTuple):
    """An input field resolved from a schema coordinate."""

    type: GraphQLInputObjectType
    input_field: GraphQLInputField
    kind: str = "InputField"


class ResolvedEnumValue(NamedTuple):
    """An enum value resolved from a schema coordinate."""

    type: GraphQLEnumType
    enum_value: GraphQLEnumValue
    kind: str = "EnumValue"


class ResolvedFieldArgument(NamedTuple):
    """A field argument resolved from a schema coordinate."""

    type: Union[GraphQLObjectType, GraphQLInterfaceType]
    field: GraphQLField
    field_argument: GraphQLArgument
    kind: str = "FieldArgument"


class ResolvedDirective(NamedTuple):
    """A directive resolved from a schema coordinate."""

    directive: GraphQLDirective
    kind: str = "Directive"


class ResolvedDirectiveArgument(NamedTuple):
    """A directive argument resolved from a schema coordinate."""

    directive: GraphQLDirective
    directive_argument: GraphQLArgument
    kind: str = "DirectiveArgument"


ResolvedSchemaElement = Union[
    ResolvedNamedType,
    ResolvedField,
    ResolvedInputField,
    ResolvedEnumValue,
    ResolvedFieldArgument,
    ResolvedDirective,
    ResolvedDirectiveArgument,
]


def resolve_schema_coordinate(
    schema: GraphQLSchema, schema_coordinate: Union[str, Source]
) -> Optional[ResolvedSchemaElement]:
    """Resolve a string schema coordinate in the context of a GraphQL schema.

    A schema coordinate is resolved in the context of a GraphQL schema to uniquely
    identify a schema element. It returns None if the schema coordinate does not
    resolve to a schema element, meta-field, or introspection schema element. It will
    raise an error if the containing schema element (if applicable) does not exist.

    `<https://spec.graphql.org/draft/#sec-Schema-Coordinates.Semantics>`_
    """
    return resolve_ast_schema_coordinate(
        schema, parse_schema_coordinate(schema_coordinate)
    )


def resolve_type_coordinate(
    schema: GraphQLSchema, schema_coordinate: TypeCoordinateNode
) -> Optional[ResolvedNamedType]:
    """TypeCoordinate : Name"""
    # 1. Let {typeName} be the value of {Name}.
    type_name = schema_coordinate.name.value
    type_ = schema.get_type(type_name)

    # 2. Return the type in the {schema} named {typeName} if it exists.
    if type_ is None:
        return None

    return ResolvedNamedType(type_)


def resolve_member_coordinate(
    schema: GraphQLSchema, schema_coordinate: MemberCoordinateNode
) -> Optional[Union[ResolvedField, ResolvedInputField, ResolvedEnumValue]]:
    """MemberCoordinate : Name . Name"""
    # 1. Let {typeName} be the value of the first {Name}.
    # 2. Let {type} be the type in the {schema} named {typeName}.
    type_name = schema_coordinate.name.value
    type_ = schema.get_type(type_name)

    # 3. Assert: {type} must exist, and must be an Enum, Input Object, Object or
    #    Interface type.
    if type_ is None:
        raise TypeError(
            f"Expected {inspect(type_name)} to be defined as a type in the schema."
        )
    if not isinstance(
        type_,
        (
            GraphQLEnumType,
            GraphQLInputObjectType,
            GraphQLObjectType,
            GraphQLInterfaceType,
        ),
    ):
        raise TypeError(
            f"Expected {inspect(type_name)}"
            " to be an Enum, Input Object, Object or Interface type."
        )

    member_name = schema_coordinate.member_name.value

    # 4. If {type} is an Enum type:
    if isinstance(type_, GraphQLEnumType):
        # 1. Let {enumValueName} be the value of the second {Name}.
        # 2. Return the enum value of {type} named {enumValueName} if it exists.
        enum_value = type_.values.get(member_name)
        if enum_value is None:
            return None
        return ResolvedEnumValue(type_, enum_value)

    # 5. Otherwise, if {type} is an Input Object type:
    if isinstance(type_, GraphQLInputObjectType):
        # 1. Let {inputFieldName} be the value of the second {Name}.
        # 2. Return the input field of {type} named {inputFieldName} if it exists.
        input_field = type_.fields.get(member_name)
        if input_field is None:
            return None
        return ResolvedInputField(type_, input_field)

    # 6. Otherwise:
    # 1. Let {fieldName} be the value of the second {Name}.
    # 2. Return the field of {type} named {fieldName} if it exists.
    field = type_.fields.get(member_name)
    if field is None:
        return None
    return ResolvedField(type_, field)


def resolve_argument_coordinate(
    schema: GraphQLSchema, schema_coordinate: ArgumentCoordinateNode
) -> Optional[ResolvedFieldArgument]:
    """ArgumentCoordinate : Name . Name ( Name : )"""
    # 1. Let {typeName} be the value of the first {Name}.
    # 2. Let {type} be the type in the {schema} named {typeName}.
    type_name = schema_coordinate.name.value
    type_ = schema.get_type(type_name)

    # 3. Assert: {type} must exist, and be an Object or Interface type.
    if type_ is None:
        raise TypeError(
            f"Expected {inspect(type_name)} to be defined as a type in the schema."
        )
    if not isinstance(type_, (GraphQLObjectType, GraphQLInterfaceType)):
        raise TypeError(
            f"Expected {inspect(type_name)} to be an object type or interface type."
        )

    # 4. Let {fieldName} be the value of the second {Name}.
    # 5. Let {field} be the field of {type} named {fieldName}.
    field_name = schema_coordinate.field_name.value
    field = type_.fields.get(field_name)

    # 6. Assert: {field} must exist.
    if field is None:
        raise TypeError(
            f"Expected {inspect(field_name)} to exist as a field"
            f" of type {inspect(type_name)} in the schema."
        )

    # 7. Let {fieldArgumentName} be the value of the third {Name}.
    field_argument_name = schema_coordinate.argument_name.value
    field_argument = field.args.get(field_argument_name)

    # 8. Return the argument of {field} named {fieldArgumentName} if it exists.
    if field_argument is None:
        return None

    return ResolvedFieldArgument(type_, field, field_argument)


def resolve_directive_coordinate(
    schema: GraphQLSchema, schema_coordinate: DirectiveCoordinateNode
) -> Optional[ResolvedDirective]:
    """DirectiveCoordinate : @ Name"""
    # 1. Let {directiveName} be the value of {Name}.
    directive_name = schema_coordinate.name.value
    directive = schema.get_directive(directive_name)

    # 2. Return the directive in the {schema} named {directiveName} if it exists.
    if directive is None:
        return None

    return ResolvedDirective(directive)


def resolve_directive_argument_coordinate(
    schema: GraphQLSchema, schema_coordinate: DirectiveArgumentCoordinateNode
) -> Optional[ResolvedDirectiveArgument]:
    """DirectiveArgumentCoordinate : @ Name ( Name : )"""
    # 1. Let {directiveName} be the value of the first {Name}.
    # 2. Let {directive} be the directive in the {schema} named {directiveName}.
    directive_name = schema_coordinate.name.value
    directive = schema.get_directive(directive_name)

    # 3. Assert {directive} must exist.
    if directive is None:
        raise TypeError(
            f"Expected {inspect(directive_name)}"
            " to be defined as a directive in the schema."
        )

    # 4. Let {directiveArgumentName} be the value of the second {Name}.
    directive_argument_name = schema_coordinate.argument_name.value
    directive_argument = directive.args.get(directive_argument_name)

    # 5. Return the argument of {directive} named {directiveArgumentName} if it exists.
    if directive_argument is None:
        return None

    return ResolvedDirectiveArgument(directive, directive_argument)


def resolve_ast_schema_coordinate(
    schema: GraphQLSchema, schema_coordinate: SchemaCoordinateNode
) -> Optional[ResolvedSchemaElement]:
    """Resolve schema coordinate from a parsed SchemaCoordinate node."""
    if isinstance(schema_coordinate, TypeCoordinateNode):
        return resolve_type_coordinate(schema, schema_coordinate)
    if isinstance(schema_coordinate, MemberCoordinateNode):
        return resolve_member_coordinate(schema, schema_coordinate)
    if isinstance(schema_coordinate, ArgumentCoordinateNode):
        return resolve_argument_coordinate(schema, schema_coordinate)
    if isinstance(schema_coordinate, DirectiveCoordinateNode):
        return resolve_directive_coordinate(schema, schema_coordinate)
    # DirectiveArgumentCoordinateNode is the only remaining kind.
    return resolve_directive_argument_coordinate(schema, schema_coordinate)



from typing import Any, Callable, List, Optional, Union, cast

from ..language import (
    ArgumentNode,
    DirectiveNode,
    EnumValueNode,
    FieldNode,
    InlineFragmentNode,
    ListValueNode,
    Node,
    ObjectFieldNode,
    OperationDefinitionNode,
    SelectionSetNode,
    VariableDefinitionNode,
    Visitor,
)
from ..pyutils import Undefined
from ..type import (
    GraphQLArgument,
    GraphQLCompositeType,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLSchema,
    GraphQLType,
    is_composite_type,
    is_input_type,
    is_output_type,
    get_named_type,
    SchemaMetaFieldDef,
    TypeMetaFieldDef,
    TypeNameMetaFieldDef,
    is_object_type,
    is_interface_type,
    get_nullable_type,
    is_list_type,
    is_input_object_type,
    is_enum_type,
)
from .type_from_ast import type_from_ast

__all__ = ["TypeInfo", "TypeInfoVisitor"]


GetFieldDefFn = Callable[
    [GraphQLSchema, GraphQLType, FieldNode], Optional[GraphQLField]
]


class TypeInfo:
    """Utility class for keeping track of type definitions.

    TypeInfo is a utility class which, given a GraphQL schema, can keep track of the
    current field and type definitions at any point in a GraphQL document AST during
    a recursive descent by calling :meth:`enter(node) <.TypeInfo.enter>` and
    :meth:`leave(node) <.TypeInfo.leave>`.
    """

    def __init__(
        self,
        schema: GraphQLSchema,
        initial_type: Optional[GraphQLType] = None,
        get_field_def_fn: Optional[GetFieldDefFn] = None,
    ) -> None:
        """Initialize the TypeInfo for the given GraphQL schema.

        Initial type may be provided in rare cases to facilitate traversals beginning
        somewhere other than documents.

        The optional ``get_field_def_fn`` parameter is deprecated; omit it so that
        TypeInfo uses its built-in field definition lookup. It will be removed in v3.3.
        """
        self._schema = schema
        self._type_stack: List[Optional[GraphQLOutputType]] = []
        self._parent_type_stack: List[Optional[GraphQLCompositeType]] = []
        self._input_type_stack: List[Optional[GraphQLInputType]] = []
        self._field_def_stack: List[Optional[GraphQLField]] = []
        self._default_value_stack: List[Any] = []
        self._directive: Optional[GraphQLDirective] = None
        self._argument: Optional[GraphQLArgument] = None
        self._enum_value: Optional[GraphQLEnumValue] = None
        self._get_field_def: GetFieldDefFn = get_field_def_fn or get_field_def
        if initial_type:
            if is_input_type(initial_type):
                self._input_type_stack.append(cast(GraphQLInputType, initial_type))
            if is_composite_type(initial_type):
                self._parent_type_stack.append(cast(GraphQLCompositeType, initial_type))
            if is_output_type(initial_type):
                self._type_stack.append(cast(GraphQLOutputType, initial_type))

    def get_type(self) -> Optional[GraphQLOutputType]:
        if self._type_stack:
            return self._type_stack[-1]
        return None

    def get_parent_type(self) -> Optional[GraphQLCompositeType]:
        if self._parent_type_stack:
            return self._parent_type_stack[-1]
        return None

    def get_input_type(self) -> Optional[GraphQLInputType]:
        if self._input_type_stack:
            return self._input_type_stack[-1]
        return None

    def get_parent_input_type(self) -> Optional[GraphQLInputType]:
        if len(self._input_type_stack) > 1:
            return self._input_type_stack[-2]
        return None

    def get_field_def(self) -> Optional[GraphQLField]:
        if self._field_def_stack:
            return self._field_def_stack[-1]
        return None

    def get_default_value(self) -> Any:
        if self._default_value_stack:
            return self._default_value_stack[-1]
        return None

    def get_directive(self) -> Optional[GraphQLDirective]:
        return self._directive

    def get_argument(self) -> Optional[GraphQLArgument]:
        return self._argument

    def get_enum_value(self) -> Optional[GraphQLEnumValue]:
        return self._enum_value

    def enter(self, node: Node) -> None:
        method = getattr(self, "enter_" + node.kind, None)
        if method:
            method(node)

    def leave(self, node: Node) -> None:
        method = getattr(self, "leave_" + node.kind, None)
        if method:
            method()

    # noinspection PyUnusedLocal
    def enter_selection_set(self, node: SelectionSetNode) -> None:
        named_type = get_named_type(self.get_type())
        self._parent_type_stack.append(
            cast(GraphQLCompositeType, named_type)
            if is_composite_type(named_type)
            else None
        )

    def enter_field(self, node: FieldNode) -> None:
        parent_type = self.get_parent_type()
        if parent_type:
            field_def = self._get_field_def(self._schema, parent_type, node)
            field_type = field_def.type if field_def else None
        else:
            field_def = field_type = None
        self._field_def_stack.append(field_def)
        self._type_stack.append(field_type if is_output_type(field_type) else None)

    def enter_directive(self, node: DirectiveNode) -> None:
        self._directive = self._schema.get_directive(node.name.value)

    def enter_operation_definition(self, node: OperationDefinitionNode) -> None:
        root_type = self._schema.get_root_type(node.operation)
        self._type_stack.append(root_type if is_object_type(root_type) else None)

    def enter_inline_fragment(self, node: InlineFragmentNode) -> None:
        type_condition_ast = node.type_condition
        output_type = (
            type_from_ast(self._schema, type_condition_ast)
            if type_condition_ast
            else get_named_type(self.get_type())
        )
        self._type_stack.append(
            cast(GraphQLOutputType, output_type)
            if is_output_type(output_type)
            else None
        )

    enter_fragment_definition = enter_inline_fragment

    def enter_variable_definition(self, node: VariableDefinitionNode) -> None:
        input_type = type_from_ast(self._schema, node.type)
        self._input_type_stack.append(
            cast(GraphQLInputType, input_type) if is_input_type(input_type) else None
        )

    def enter_argument(self, node: ArgumentNode) -> None:
        field_or_directive = self.get_directive() or self.get_field_def()
        if field_or_directive:
            arg_def = field_or_directive.args.get(node.name.value)
            arg_type = arg_def.type if arg_def else None
        else:
            arg_def = arg_type = None
        self._argument = arg_def
        self._default_value_stack.append(
            arg_def.default_value if arg_def else Undefined
        )
        self._input_type_stack.append(arg_type if is_input_type(arg_type) else None)

    # noinspection PyUnusedLocal
    def enter_list_value(self, node: ListValueNode) -> None:
        list_type = get_nullable_type(self.get_input_type())  # type: ignore
        item_type = (
            cast(GraphQLList, list_type).of_type
            if is_list_type(list_type)
            else list_type
        )
        # List positions never have a default value.
        self._default_value_stack.append(Undefined)
        self._input_type_stack.append(item_type if is_input_type(item_type) else None)

    def enter_object_field(self, node: ObjectFieldNode) -> None:
        object_type = get_named_type(self.get_input_type())
        if is_input_object_type(object_type):
            input_field = cast(GraphQLInputObjectType, object_type).fields.get(
                node.name.value
            )
            input_field_type = input_field.type if input_field else None
        else:
            input_field = input_field_type = None
        self._default_value_stack.append(
            input_field.default_value if input_field else Undefined
        )
        self._input_type_stack.append(
            input_field_type if is_input_type(input_field_type) else None
        )

    def enter_enum_value(self, node: EnumValueNode) -> None:
        enum_type = get_named_type(self.get_input_type())
        if is_enum_type(enum_type):
            enum_value = cast(GraphQLEnumType, enum_type).values.get(node.value)
        else:
            enum_value = None
        self._enum_value = enum_value

    def leave_selection_set(self) -> None:
        del self._parent_type_stack[-1:]

    def leave_field(self) -> None:
        del self._field_def_stack[-1:]
        del self._type_stack[-1:]

    def leave_directive(self) -> None:
        self._directive = None

    def leave_operation_definition(self) -> None:
        del self._type_stack[-1:]

    leave_inline_fragment = leave_operation_definition
    leave_fragment_definition = leave_operation_definition

    def leave_variable_definition(self) -> None:
        del self._input_type_stack[-1:]

    def leave_argument(self) -> None:
        self._argument = None
        del self._default_value_stack[-1:]
        del self._input_type_stack[-1:]

    def leave_list_value(self) -> None:
        del self._default_value_stack[-1:]
        del self._input_type_stack[-1:]

    leave_object_field = leave_list_value

    def leave_enum_value(self) -> None:
        self._enum_value = None


def get_field_def(
    schema: GraphQLSchema, parent_type: GraphQLType, field_node: FieldNode
) -> Optional[GraphQLField]:
    """Get field definition.

    Not exactly the same as the executor's definition of
    :func:`graphql.execution.get_field_def`, in this statically evaluated environment
    we do not always have an Object type, and need to handle Interface and Union types.
    """
    name = field_node.name.value
    if name == "__schema" and schema.query_type is parent_type:
        return SchemaMetaFieldDef
    if name == "__type" and schema.query_type is parent_type:
        return TypeMetaFieldDef
    if name == "__typename" and is_composite_type(parent_type):
        return TypeNameMetaFieldDef
    if is_object_type(parent_type) or is_interface_type(parent_type):
        parent_type = cast(Union[GraphQLObjectType, GraphQLInterfaceType], parent_type)
        return parent_type.fields.get(name)
    return None


class TypeInfoVisitor(Visitor):
    """A visitor which maintains a provided TypeInfo."""

    def __init__(self, type_info: "TypeInfo", visitor: Visitor):
        super().__init__()
        self.type_info = type_info
        self.visitor = visitor

    def enter(self, node: Node, *args: Any) -> Any:
        self.type_info.enter(node)
        fn = self.visitor.get_enter_leave_for_kind(node.kind).enter
        if fn:
            result = fn(node, *args)
            if result is not None:
                self.type_info.leave(node)
                if isinstance(result, Node):
                    self.type_info.enter(result)
            return result

    def leave(self, node: Node, *args: Any) -> Any:
        fn = self.visitor.get_enter_leave_for_kind(node.kind).leave
        result = fn(node, *args) if fn else None
        self.type_info.leave(node)
        return result



"""GraphQL Language

The :mod:`graphql.language` package is responsible for parsing and operating on the
GraphQL language.
"""

from .source import Source

from .location import get_location, SourceLocation, FormattedSourceLocation

from .print_location import print_location, print_source_location

from .token_kind import TokenKind

from .lexer import Lexer

from .parser import (
    parse,
    parse_type,
    parse_value,
    parse_const_value,
    parse_schema_coordinate,
)

from .printer import print_ast

from .visitor import (
    visit,
    Visitor,
    ParallelVisitor,
    VisitorAction,
    VisitorKeyMap,
    BREAK,
    SKIP,
    REMOVE,
    IDLE,
)

from .ast import (
    Location,
    Token,
    Node,
    # Each kind of AST node
    NameNode,
    DocumentNode,
    DefinitionNode,
    ExecutableDefinitionNode,
    OperationDefinitionNode,
    OperationType,
    VariableDefinitionNode,
    VariableNode,
    SelectionSetNode,
    SelectionNode,
    FieldNode,
    ArgumentNode,
    ConstArgumentNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    FragmentDefinitionNode,
    ValueNode,
    ConstValueNode,
    IntValueNode,
    FloatValueNode,
    StringValueNode,
    BooleanValueNode,
    NullValueNode,
    EnumValueNode,
    ListValueNode,
    ConstListValueNode,
    ObjectValueNode,
    ConstObjectValueNode,
    ObjectFieldNode,
    ConstObjectFieldNode,
    DirectiveNode,
    ConstDirectiveNode,
    TypeNode,
    NamedTypeNode,
    ListTypeNode,
    NonNullTypeNode,
    TypeSystemDefinitionNode,
    SchemaDefinitionNode,
    OperationTypeDefinitionNode,
    TypeDefinitionNode,
    ScalarTypeDefinitionNode,
    ObjectTypeDefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    UnionTypeDefinitionNode,
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    InputObjectTypeDefinitionNode,
    DirectiveDefinitionNode,
    TypeSystemExtensionNode,
    SchemaExtensionNode,
    TypeExtensionNode,
    ScalarTypeExtensionNode,
    ObjectTypeExtensionNode,
    InterfaceTypeExtensionNode,
    UnionTypeExtensionNode,
    EnumTypeExtensionNode,
    InputObjectTypeExtensionNode,
    DirectiveExtensionNode,
    SchemaCoordinateNode,
    TypeCoordinateNode,
    MemberCoordinateNode,
    ArgumentCoordinateNode,
    DirectiveCoordinateNode,
    DirectiveArgumentCoordinateNode,
)
from .predicates import (
    is_definition_node,
    is_executable_definition_node,
    is_selection_node,
    is_value_node,
    is_const_value_node,
    is_type_node,
    is_type_system_definition_node,
    is_type_definition_node,
    is_type_system_extension_node,
    is_type_extension_node,
    is_schema_coordinate_node,
)
from .directive_locations import DirectiveLocation

__all__ = [
    "get_location",
    "SourceLocation",
    "FormattedSourceLocation",
    "print_location",
    "print_source_location",
    "TokenKind",
    "Lexer",
    "parse",
    "parse_value",
    "parse_const_value",
    "parse_type",
    "parse_schema_coordinate",
    "print_ast",
    "Source",
    "visit",
    "Visitor",
    "ParallelVisitor",
    "VisitorAction",
    "VisitorKeyMap",
    "BREAK",
    "SKIP",
    "REMOVE",
    "IDLE",
    "Location",
    "Token",
    "DirectiveLocation",
    "Node",
    "NameNode",
    "DocumentNode",
    "DefinitionNode",
    "ExecutableDefinitionNode",
    "OperationDefinitionNode",
    "OperationType",
    "VariableDefinitionNode",
    "VariableNode",
    "SelectionSetNode",
    "SelectionNode",
    "FieldNode",
    "ArgumentNode",
    "ConstArgumentNode",
    "FragmentSpreadNode",
    "InlineFragmentNode",
    "FragmentDefinitionNode",
    "ValueNode",
    "ConstValueNode",
    "IntValueNode",
    "FloatValueNode",
    "StringValueNode",
    "BooleanValueNode",
    "NullValueNode",
    "EnumValueNode",
    "ListValueNode",
    "ConstListValueNode",
    "ObjectValueNode",
    "ConstObjectValueNode",
    "ObjectFieldNode",
    "ConstObjectFieldNode",
    "DirectiveNode",
    "ConstDirectiveNode",
    "TypeNode",
    "NamedTypeNode",
    "ListTypeNode",
    "NonNullTypeNode",
    "TypeSystemDefinitionNode",
    "SchemaDefinitionNode",
    "OperationTypeDefinitionNode",
    "TypeDefinitionNode",
    "ScalarTypeDefinitionNode",
    "ObjectTypeDefinitionNode",
    "FieldDefinitionNode",
    "InputValueDefinitionNode",
    "InterfaceTypeDefinitionNode",
    "UnionTypeDefinitionNode",
    "EnumTypeDefinitionNode",
    "EnumValueDefinitionNode",
    "InputObjectTypeDefinitionNode",
    "DirectiveDefinitionNode",
    "TypeSystemExtensionNode",
    "SchemaExtensionNode",
    "TypeExtensionNode",
    "ScalarTypeExtensionNode",
    "ObjectTypeExtensionNode",
    "InterfaceTypeExtensionNode",
    "UnionTypeExtensionNode",
    "EnumTypeExtensionNode",
    "InputObjectTypeExtensionNode",
    "DirectiveExtensionNode",
    "SchemaCoordinateNode",
    "TypeCoordinateNode",
    "MemberCoordinateNode",
    "ArgumentCoordinateNode",
    "DirectiveCoordinateNode",
    "DirectiveArgumentCoordinateNode",
    "is_definition_node",
    "is_executable_definition_node",
    "is_selection_node",
    "is_value_node",
    "is_const_value_node",
    "is_type_node",
    "is_type_system_definition_node",
    "is_type_definition_node",
    "is_type_system_extension_node",
    "is_type_extension_node",
    "is_schema_coordinate_node",
]
